import os


def mkdir_dataset(new_dataset_folder_name):
    new_dataset_folder_path = f'{os.getcwd()}/dataset/{new_dataset_folder_name}'
    if not os.path.exists(new_dataset_folder_path):
        os.mkdir(new_dataset_folder_path)
        print('dataset folder created inside dataset: ' + new_dataset_folder_name)
    else:
        print('dataset folder already exists: ' + new_dataset_folder_name)


def mkdir_output():
    output_folder_path = f'{os.getcwd()}/output'

    try:
        if not os.path.exists(output_folder_path):
            os.mkdir(output_folder_path)

    except FileExistsError:
        pass


def get_directory_name(name, x=0):
    dir_name = (name + (' ' + str(x) if x != 0 else '')).strip()
    if not os.path.exists(dir_name):
        return dir_name
    else:
        return get_directory_name(name, x + 1)



import pandas as pd
from keras.preprocessing.image import ImageDataGenerator


def get_train_generator(x, y, noise, train_index, batch_size, class_mode, input_shape):
    x = list(map(lambda p: p.replace('default', f'noise_{noise}'), x))
    df = pd.DataFrame({'images': x})
    x = df['images']

    x_train, y_train = x[train_index], y[train_index]

    df_train = pd.DataFrame({'id': x_train, 'label': y_train})

    train_gen = ImageDataGenerator(
        rotation_range=40,
        rescale=1 / 255,
        horizontal_flip=True,
        vertical_flip=True,
        validation_split=0.1
    )

    train_generator = train_gen.flow_from_dataframe(
        dataframe=df_train,
        x_col='id',
        y_col='label',
        batch_size=batch_size,
        target_size=input_shape,
        class_mode=class_mode,
        shuffle=False,
        subset='training')

    validation_generator = train_gen.flow_from_dataframe(
        dataframe=df_train,
        x_col='id',
        y_col='label',
        batch_size=batch_size,
        target_size=input_shape,
        class_mode=class_mode,
        shuffle=False,
        subset='validation')

    return train_generator, validation_generator


def get_test_generator(x, y, noise, test_index, batch_size, class_mode, input_shape):
    x = list(map(lambda p: p.replace('default', f'noise_{noise}'), x))
    df = pd.DataFrame({'images': x})
    x = df['images']

    x_test, y_test = x[test_index], y[test_index]

    df_test = pd.DataFrame({'id': x_test, 'label': y_test})

    test_generator = ImageDataGenerator(
        rescale=1 / 255,
    ).flow_from_dataframe(
        dataframe=df_test,
        x_col='id',
        y_col='label',
        batch_size=batch_size,
        target_size=input_shape,
        class_mode=class_mode,
        shuffle=False)

    return test_generator



import os

import cv2
import numpy as np
from skimage.util import random_noise

from ..utils.mkdir_folders import mkdir_dataset


def __write_images(dataset_name, noise_amount, image_path_arr):
    print(f'Generating with noise of {noise_amount} for {os.path.basename(image_path_arr)}')
    img = cv2.imread(image_path_arr)
    new_image_path = f'{os.getcwd()}/dataset/{dataset_name}/{os.path.basename(image_path_arr)}'
    if not os.path.exists(new_image_path):
        gray_image = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        noise_img = random_noise(gray_image, mode='s&p', amount=noise_amount)
        noise_img = np.array(255 * noise_img, dtype='uint8')

        cv2.imwrite(new_image_path, noise_img)


def generate_dataset(x_data, dataset_name, noise_amount):
    mkdir_dataset(dataset_name)
    print('Copying images to the new dataset')
    [
        __write_images(dataset_name, noise_amount, path_array)
        for path_array in x_data
    ]



from keras.models import Sequential


# def __create_callbacks(alpha):
#     lr_reduce = ReduceLROnPlateau(monitor='val_loss', factor=0.1, min_delta=alpha, patience=5, verbose=1)
#     erl_stopping = tf.keras.callbacks.EarlyStopping(patience=3, monitor='val_loss', verbose=1)
#     callbacks = [lr_reduce, erl_stopping]
#     return callbacks

def train_model_for_dataset(model: Sequential, train_generator, validation_generator):
    # callbacks = __create_callbacks(alpha)

    history = model.fit(
        train_generator,
        validation_data=validation_generator,
        # callbacks=callbacks,
        epochs=1
    )

    return history



def isfloat(num):
    try:
        float(num)
        return True
    except ValueError:
        return False


def get_results_columns(target_names):
    classes_precision_columns = []
    classes_recall_columns = []
    classes_f1score_columns = []
    for index, item in enumerate(target_names):
        classes_precision_columns.append(f'precision({item})')
        classes_recall_columns.append(f'recall({item})')
        classes_f1score_columns.append(f'f1-score({item})')

    return [
        'train-dataset-noise', 'test-dataset-noise',
        *classes_precision_columns,
        'precision(macro-avg)', 'precision(weighted-avg)',
        *classes_recall_columns,
        'recall(macro-avg)',  'recall(weighted-avg)',
        *classes_f1score_columns,
        'f1-score(accuracy)', 'f1-score(macro-avg)', 'f1-score(weighted-avg)'
    ]


import os
import re
from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from .functions import get_results_columns
from .mkdir_folders import mkdir_output, get_directory_name


class ResultsWriter:

    def __init__(self, name):
        mkdir_output()
        self.execution_folder_path = get_directory_name(f'{os.getcwd()}/output/{name}')
        os.mkdir(self.execution_folder_path)
        self.results_folder = '/'
        self.results_name = name

    def __generate_df_by_csv(self):
        df = pd.read_csv(self.results_folder + f'/results_{self.results_name}.csv')
        df.drop('Unnamed: 0', axis=1, inplace=True)
        return df

    def __generate_results_csv(self, target_names):
        pd.DataFrame(columns=get_results_columns(target_names)).to_csv(self.results_folder + f'/results_{self.results_name}.csv')

    def write_metrics_results(self, train_noise, test_noise, cr, cm, target_names):
        if not os.path.isfile(self.results_folder + f'/results_{self.results_name}.csv'):
            self.__generate_results_csv(target_names)

        current_df = self.__generate_df_by_csv()

        pattern = re.compile(r'(\w+)\s+([\d.]+)\s+([\d.]+)\s+([\d.]+)\s+(\d+)')
        matches = pattern.findall(cr)

        columns = ['Class', 'Precision', 'Recall', 'F1-Score', 'Support']
        df = pd.DataFrame(matches, columns=columns).astype(
            {'Precision': float, 'Recall': float, 'F1-Score': float, 'Support': int})

        accuracy_match = re.search(r'accuracy\s+([\d.]+)', cr)
        accuracy = float(accuracy_match.group(1)) if accuracy_match else None

        classes_precision_dict = {}
        classes_recall_dict = {}
        classes_f1score_dict = {}
        for index, item in enumerate(target_names):
            classes_precision_dict[f'precision({item})'] = df['Precision'][index]
            classes_recall_dict[f'recall({item})'] = df['Recall'][index]
            classes_f1score_dict[f'f1-score({item})'] = df['F1-Score'][index]

        classes_number = len(target_names)
        metrics = {
            'train-dataset-noise': train_noise,
            'test-dataset-noise': test_noise,
            **classes_precision_dict,
            'precision(macro-avg)': df['Precision'][classes_number],
            'precision(weighted-avg)': df['Precision'][classes_number + 1],
            **classes_recall_dict,
            'recall(macro-avg)': df['Recall'][classes_number],
            'recall(weighted-avg)': df['Recall'][classes_number + 1],
            **classes_f1score_dict,
            'f1-score(accuracy)': accuracy,
            'f1-score(macro-avg)': df['F1-Score'][classes_number],
            'f1-score(weighted-avg)': df['F1-Score'][classes_number + 1],
        }

        pd.DataFrame(cm).to_csv(
            self.results_folder + f'/train_{train_noise}_test_{test_noise}_confusion_matrix.csv'
        )
        pd.concat([current_df, pd.DataFrame(metrics, index=[0])], ignore_index=True) \
            .to_csv(self.results_folder + f'/results_{self.results_name}.csv')

    def write_execution_folder(self):
        self.results_folder = f'{self.execution_folder_path}/results_{self.results_name}_{datetime.now().isoformat().__str__()}'
        os.mkdir(self.results_folder)

    def write_model(self, model, model_name):
        model.save(f'{self.results_folder}/{model_name}.hdf5')

    def delete_results(self):
        os.rmdir(self.results_folder)

    def generate_mean_csv(self):
        sns.set_theme(style="white")
        unified_data = []
        train_noise = []
        test_noise = []
        for subdir, _, files in os.walk(f'{os.getcwd()}/output/{self.results_name}'):
            for file in files:
                if file.endswith('.csv') and file.startswith('results'):
                    file_path = os.path.join(subdir, file)
                    csv = pd.read_csv(file_path)
                    unified_data.append(csv['f1-score(weighted-avg)'])
                    train_noise = csv['train-dataset-noise']
                    test_noise = csv['test-dataset-noise']

        merged_df = pd.concat(unified_data, axis=1)
        mean_data = {'train-noise': train_noise, 'test-noise': test_noise,
                     'f1-score(weighted-avg)': merged_df.mean(axis=1)}

        mean_df = pd.DataFrame(mean_data)

        mean_df.to_csv(
            f'{os.getcwd()}/output/{self.results_name}/mean_results.csv')

        column_values = mean_df['train-noise'].unique()
        conf_matrix = []

        for value in column_values:
            value_df = mean_df[mean_df['train-noise'] == value]
            conf_matrix.append(value_df['f1-score(weighted-avg)'].to_numpy())

        heatmap_df = pd.DataFrame(conf_matrix, columns=column_values, index=column_values)

        sns.set_theme(style="whitegrid")
        sns.set(font_scale=2)
        sns.set_context("paper")
        sns.set(font='serif')
        sns.set_style("white", {
            "font.family": "serif",
            "font.serif": ["Times", "Palatino", "serif"]
        })

        plt.figure(figsize=(8, 6))
        sns.heatmap(heatmap_df,
                    cmap='coolwarm',
                    annot=True,
                    cbar_kws={'label': '5-Fold F-Score mean vs. 0 Test Noise'},
                    fmt='.2f')

        plt.xlabel('Test Noise')
        plt.ylabel('Train Noise')

        plt.savefig(f'{os.getcwd()}/output/{self.results_name}/mean_results_heatmap.png')
